
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <limits.h>

#include "2pc.h"

#ifdef USE_DC
#include "dc.h" 
#endif


/**
 ** Globals
 **/

char		*appname;
int		nprocs = 0;
commit_state	states[MAX_TRANS];
int		curr_trans = 0;
u_long		hosts[MAX_HOSTS];
int		ports[MAX_HOSTS];
int		num_deferred_regs = 0;
struct sockaddr_in deferred_regs[MAX_DEFERRED];
int		num_deferred_deregs = 0;
struct sockaddr_in deferred_deregs[MAX_DEFERRED];


/**
 ** Functions
 **/

void pmsg(char* msg)
{
	fprintf(stderr, "%s: %s\n", appname, msg);
}

void perr(char* msg)
{
	char	buf[128];
	sprintf(buf, "%s: %s", appname, msg);
	perror(buf);
}


/* 
 * get_address() converts its host argument string (either a hostname or an
 * IP address in dotted decimal format), into an internet address and returns
 * it, or FALSE on failure.
 */ 
unsigned long get_address(char* host)
{
	struct in_addr  *addr_ptr;
	unsigned long   inaddr;

	/*
	 * First check if host string we've been passed is in
	 * dotted decimal format. Failing that, we'll get the
	 * IP from the system. 
	 */     
	if ((int)(inaddr = inet_addr(host)) == INADDR_NONE) {
		struct hostent  *hp;
		struct in_addr  *addr;

		hp = gethostbyname(host);
		if (hp == NULL) {
			return FALSE;
		}
		addr = (struct in_addr*) hp->h_addr;
		return addr->s_addr;
	}
	else {
		return inaddr;
	}
}


int udp_send(int sock, void* data, int len, int flags, u_long ip, int port)
{  
	int             res;
	struct timeval  t = {0, 1}; /* 1 us */
	struct sockaddr_in d;

	/* Fill in destination sockaddr struct */
	d.sin_family            = AF_INET;
	d.sin_port              = htons(port);
	d.sin_addr.s_addr       = ip;

	/*
	 * The Freebsd messaging code easily fills up its available
	 * message buffers (~50 of them). If there are no available buffers
	 * we'll sleep for 1 clock tick and then try again.
	 */
	while ((res = sendto(sock, data, len, flags, (struct sockaddr*) &d,
		sizeof(d))) < 0 && errno == ENOBUFS) {
		select(0, NULL, NULL, NULL, &t);
	}
	return res;
}


int udp_recv(int sock, void* b, int len, int flags, struct sockaddr_in* addr)
{
	int             val, nlen;

	nlen = (addr == NULL ? 0 : sizeof(struct sockaddr_in));

	/*
	 * Retry this receive if it was interrupted by a signal
	 */
	while ((val = recvfrom(sock, b, len, flags, (struct sockaddr*)addr,
	       &nlen)) < 0 && errno == EINTR);
	return val;
}        

int ip2proc(struct sockaddr_in *addr)
{
	int	i;
	u_long	ip = addr->sin_addr.s_addr;

	for (i=0; i < nprocs; i++) {
		if (hosts[i] == ip) {
			ports[i] = ntohs(addr->sin_port);
			return i;
		}
	}
	pmsg("got msg from unknown ip");
	fprintf(stderr, "ip = 0x%x\n", ip);
	return(-1);
}

/*
 * Subtract time t2 from time t1 and store the result in dst. If t2 is
 * bigger than t1, store 0 in dst. 
 */
void sub_time(struct timeval* t1, struct timeval* t2, struct timeval* dst)
{
        struct timeval  x, y;

        x = *t1;
        y = *t2;

        /* check that operands can be safely subtracted */
        if (y.tv_sec>x.tv_sec || (y.tv_sec==x.tv_sec && y.tv_usec>x.tv_usec)) {
                dst->tv_sec = 0;
                dst->tv_usec = 0;
                return; 
        }

        /* compute x - y */
        if (y.tv_usec > x.tv_usec) {
                /* borrow from seconds, give to usecs */
                x.tv_sec--;
                x.tv_usec += 1000000;
        }
        dst->tv_sec = x.tv_sec - y.tv_sec;
        dst->tv_usec = x.tv_usec - y.tv_usec;
}


void do_protocol(int s, msg pm, struct sockaddr_in *cli)
{
	int	proc, i, havevoted, res;
	msg	m;
	char	*vote, buf[128];
	struct sockaddr_in c;
	struct timeval start, curr;

	if (cli != NULL) {
		proc = ip2proc(cli);

		/*
		 * Drop messages from unknown processes (e.g. an old COMMIT
		 * message from a process that de-registered). ???
		 */
		if (proc < 0) {
		    return;
		}
		states[curr_trans].votes[proc] = pm.aborted ? ABORT : COMMIT;
		sprintf(buf, "process %d voted %s", proc, pm.aborted?"ABORT":"COMMIT");
		pmsg(buf);
	}

	/*
	 * PHASE 1
	 *
	 * Send out messages to tell each process to PREPARE, and get
	 * back their PREPARED + vote msg. If we don't hear back from 
	 * each process in a timely fashion, send out a new round of 
	 * PREPARE msgs to the pokey processes.
	 */

	m.id = curr_trans;
	for (;;) {
		/*
		 * Send messages to each SIGILL daemon for each process
		 * that has yet to vote. If all processes have already voted, 
		 * note that fact and go to next phase.
		 */
		m.type = DO_PREPARE;
		havevoted = 0;
		for (i=0; i < nprocs; i++) {
			if (states[curr_trans].votes[i] != UNKNOWN) {
				havevoted++;
				continue;
			}
			res = udp_send(s, &m, sizeof(m), 0, hosts[i], SIG_PORT);
			if (res < 0) {
				perr("udp_send");
				exit(1);
			}
		}
		if (havevoted == nprocs)
			break;

		/*
		 * Note time at start of this loop so we can keep
		 * track of total time blocked over multiple
		 * select calls.
		 */
		if (gettimeofday(&start, NULL) < 0) {
			perr("gettimeofday");
			exit(1);
		}

		/*
		 * Get votes back
		 */
		while (havevoted < nprocs) {
			struct timeval diff, t = {TIMEOUT, 0};
			fd_set	fds;

			/*
			 * Subtract any used time
			 */
			if (gettimeofday(&curr, NULL) < 0) {
				perr("gettimeofday");
				exit(1);
			}
			sub_time(&curr, &start, &diff);
			sub_time(&t, &diff, &t);

			FD_ZERO(&fds);
			FD_SET(s, &fds);
			res = select(64, &fds, NULL, NULL, &t);
			if (res < 0) {
				perr("select");
				exit(1);
			}
			else if (res == 0) break;
			else {
				res = udp_recv(s, &m, sizeof(m), 0, &c);
				if (res < 0) {
					perr("udp_recv");
					exit(1);
				}

				/*
				fprintf(stderr, "\tgot from %d: type == %s, id == %d, aborted == %s\n", ip2proc(&c), (m.type==PREPARED?"PREPARED":"WEIRD"), m.id, (m.aborted?"TRUE":"FALSE"));
				*/

				if (m.type == REGISTER) {
					/*
					 * If this is a register message, defer handling 
					 * it until after we're out of the protocol.
					 */
					deferred_regs[num_deferred_regs++] = c;
					continue;
				}
				else if (m.type == DEREGISTER) {
					/*
					 * If this is a deregister message, defer handling 
					 * it until after we're out of the protocol.
					 */
					deferred_deregs[num_deferred_deregs++] = c;
					continue;
				}
				else if (m.id < 0 || m.type != PREPARED) {
					pmsg("got mad msg");
					exit(1);
				}
				if (m.id < curr_trans) {
					/*
					 * Send back prior result
					 */
					sprintf(buf, "resending %d result to process %d", m.id, ip2proc(&c));
					pmsg(buf);
					m.type = states[m.id].final_vote;
					m.new_id = curr_trans;
					res = sendto(s, &m, sizeof(m), 0, (struct sockaddr*)&c, sizeof(c));
					if (res < 0) {
						perr("sendto");
						exit(1);
					}
					continue;
				}
				/* 
				 * Drop prepare msgs that have traveled
				 * backwards in time... (should never happen)
				 */
				if (m.id > curr_trans) continue;

				proc = ip2proc(&c);
				vote = &(states[curr_trans].votes[proc]);
				if (*vote == UNKNOWN) {
					*vote = m.aborted ? ABORT : COMMIT;
					havevoted++;
				}
				sprintf(buf, "process %d voted %s", proc, m.aborted?"ABORT":"COMMIT");
				pmsg(buf);
			}
		}
		if (havevoted == nprocs)
			break;
	}

	/*
	 * PHASE 2
	 *
	 * Everyone has prepared and voted successfully. Now count up 
	 * the votes and let everyone know the result.
	 */

	for (i=0; i < nprocs; i++) {
		vote = &(states[curr_trans].votes[i]);
		if (*vote == ABORT) break;
		else if (*vote == COMMIT) continue;
		else {
			pmsg("recorded vote had non COMMIT/ABORT value");
			exit(1);
		}
	}
	states[curr_trans].final_vote = (i == nprocs ? COMMIT : ABORT);

	sprintf(buf, "vote for trans %d was %s",curr_trans,states[curr_trans].final_vote==COMMIT?"COMMIT":"ABORT");
	pmsg(buf);

	pmsg("notifying processes of result");
	m.id = curr_trans;
	m.type = states[curr_trans].final_vote;

	curr_trans = (curr_trans + 1)%MAX_TRANS;
	bzero((char *) &states[curr_trans], sizeof(commit_state));

	m.new_id = curr_trans;

	for (i=0; i < nprocs; i++) {
		res = udp_send(s, &m, sizeof(m), 0, hosts[i], ports[i]);
		if (res < 0) {
			perr("udp_send");
			exit(1);
		}
	}
}

void do_register(int sock, struct sockaddr_in *cli)
{
	int	res, i, port;
	u_long	ip;
	msg	m;

	ip = cli->sin_addr.s_addr;
	port = ntohs(cli->sin_port);

	/*
	 * Check to make sure this process is not
	 * already registered...
	 */
	for (i=0; i < nprocs; i++) {
		if (hosts[i] == ip && ports[i] == port)
			break;
	}

	if (i == nprocs) {
		fprintf(stderr, "%s: registering process %d\n", appname, nprocs);
		nprocs++;
	}

	hosts[i] = ip;
	ports[i] = port;
	m.type 		= REGISTERED;
	m.id 		= i;
	m.new_id	= curr_trans;
	res = sendto(sock, &m, sizeof(m), 0, (struct sockaddr*)cli, sizeof(*cli));
	if (res < 0) {
		perr("sendto");
		exit(1);
	}
}

void do_deregister(int sock, struct sockaddr_in *cli)
{
	int	res, i;
	u_long	ip;
	msg	m;
	int proc;

	ip = cli->sin_addr.s_addr;

	/*
	 * Find in process table
	 */
	for (proc = 0; proc < nprocs && hosts[proc] != ip; proc++) {
	}

	if (proc < nprocs) {
		fprintf(stderr, "%s: de-registering process %d\n", appname,
			    proc);
		/*
		 * Move last entry into deleted slot
		 */
		hosts[proc] = hosts[nprocs-1];
		ports[proc] = ports[nprocs-1];
		nprocs--;
	}

	m.type 		= DEREGISTERED;
	m.id 		= i;
	m.new_id	= curr_trans;
	res = sendto(sock, &m, sizeof(m), 0, (struct sockaddr*)cli, sizeof(*cli));
	if (res < 0) {
		perr("sendto");
		exit(1);
	}
}


int main(int argc, char** argv)
{
	int	sock, res, i;
	fd_set	fds;
	char	buf[128];
	struct sockaddr_in name;

#ifdef USE_DC
	dc_init("2pcd", NULL, NULL, &argv, NULL);
#endif

	appname = argv[0];
	if (argc != 1) {
		fprintf(stderr, "Usage: %s\n", appname);
		exit(1);
	}

	/*
	 * Set up our system
	 */
	pmsg("initializing");

	bzero((char*)states, MAX_TRANS * sizeof(commit_state));

	sock = socket(PF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perr("rm_socket()");
		exit(1);
	}

	bzero(&name, sizeof(name));
	name.sin_family 	= AF_INET;
	name.sin_addr.s_addr 	= htonl(INADDR_ANY);
	name.sin_port 		= htons(SERVER_PORT);

	res = bind(sock, (struct sockaddr*)&name, sizeof(name));
	if (res < 0) {
		perr("bind");
		exit(1);
	}

	pmsg("ready");

	for (;;) {
		int	id;
		msg	m;
		struct sockaddr_in cli;

		res = udp_recv(sock, &m, sizeof(m), 0, &cli);
		if (res < 0) {
			perr("udp_recv");
			exit(1);
		}
		/*
		fprintf(stderr, "\tgot from %d: type == %s, id == %d, aborted == %s\n", ip2proc(&cli), (m.type==PREPARED?"PREPARED":"WEIRD"), m.id, (m.aborted?"TRUE":"FALSE"));
		*/

		if (m.type == REGISTER) {
			do_register(sock, &cli);
			continue;
		}
		else if (m.type == DEREGISTER) {
			do_deregister(sock, &cli);
			continue;
		}
		else if (m.type != PREPARED) {
			pmsg("got bad message type");
			exit(1);
		}

		id = m.id;
		if (states[id].final_vote != UNKNOWN) {
			m.type = states[id].final_vote;
			m.id = id;
			m.new_id = curr_trans;
			sprintf(buf, "resending vote %d result to process %d",
				id, ip2proc(&cli));
			pmsg(buf);
			res = sendto(sock, &m, sizeof(m), 0, 
			      (struct sockaddr*)&cli, sizeof(cli));
			if (res < 0) {
				perr("sendto");
				exit(1);
			}
		}
		else if (m.id != curr_trans) {
			/*
			 * Spurrious message
			 */
			sprintf(buf, "got PREPARED %d while waiting to start %d", m.id, curr_trans);
			pmsg(buf);
		}
		else {
			sprintf(buf, "starting commit prototcol for trans %d", curr_trans);
			pmsg(buf);
			do_protocol(sock, m, &cli);
			pmsg("protocol finished\n");
			/*
			 * Out of protocol, so we can handle any registers and deregisters
			 * that came in during the protocol
			 */
			for (i=0; i < num_deferred_regs; i++) {
				do_register(sock, &deferred_regs[i]);
			}
			num_deferred_regs = 0;
			for (i=0; i < num_deferred_deregs; i++) {
				do_deregister(sock, &deferred_deregs[i]);
			}
			num_deferred_deregs = 0;
		}
	}
}

